@@ -64,17 +64,15 @@ class OAuth2Session(requests.Session):
         :param kwargs: Arguments to pass to the Session constructor.
         """
         super(OAuth2Session, self).__init__(**kwargs)
-        self.client_id = client_id or client.client_id
+        self._client = client or WebApplicationClient(client_id, token=token)
+        self.token = token or {}
         self.scope = scope
         self.redirect_uri = redirect_uri
-        self.token = token or {}
         self.state = state or generate_token
         self._state = state
         self.auto_refresh_url = auto_refresh_url
         self.auto_refresh_kwargs = auto_refresh_kwargs or {}
         self.token_updater = token_updater
-        self._client = client or WebApplicationClient(client_id, token=token)
-        self._client._populate_attributes(token or {})
 
         # Allow customizations for non compliant providers through various
         # hooks to adjust requests and responses.
@@ -94,6 +92,50 @@ class OAuth2Session(requests.Session):
             log.debug('Re-using previously supplied state %s.', self._state)
         return self._state
 
+    @property
+    def client_id(self):
+        return getattr(self._client, "client_id", None)
+
+    @client_id.setter
+    def client_id(self, value):
+        self._client.client_id = value
+
+    @client_id.deleter
+    def client_id(self):
+        del self._client.client_id
+
+    @property
+    def token(self):
+        return getattr(self._client, "token", None)
+
+    @token.setter
+    def token(self, value):
+        self._client.token = value
+        self._client._populate_attributes(value)
+
+    @property
+    def access_token(self):
+        return getattr(self._client, "access_token", None)
+
+    @access_token.setter
+    def access_token(self, value):
+        self._client.access_token = value
+
+    @access_token.deleter
+    def access_token(self):
+        del self._client.access_token
+
+    @property
+    def authorized(self):
+        """Boolean that indicates whether this session has an OAuth token
+        or not. If `self.authorized` is True, you can reasonably expect
+        OAuth-protected requests to the resource to succeed. If
+        `self.authorized` is False, you need the user to go through the OAuth
+        authentication dance before OAuth-protected requests to the resource
+        will succeed.
+        """
+        return bool(self.access_token)
+
     def authorization_url(self, url, state=None, **kwargs):
         """Form an authorization URL.
 
@@ -112,7 +154,7 @@ class OAuth2Session(requests.Session):
 
     def fetch_token(self, token_url, code=None, authorization_response=None,
             body='', auth=None, username=None, password=None, method='POST',
-            verify=True, **kwargs):
+            timeout=None, headers=None, verify=True, proxies=None, **kwargs):
         """Generic method for fetching an access token from the token endpoint.
 
         If you are using the MobileApplicationClient you will want to use
@@ -131,6 +173,8 @@ class OAuth2Session(requests.Session):
         :param method: The HTTP method used to make the request. Defaults
                        to POST, but may also be GET. Other methods should
                        be added as needed.
+        :param headers: Dict to default request headers with.
+        :param timeout: Timeout of the request in seconds.
         :param verify: Verify SSL certificate.
         :param kwargs: Extra parameters to include in the token request.
         :return: A token dict
@@ -146,29 +190,49 @@ class OAuth2Session(requests.Session):
             code = self._client.code
             if not code:
                 raise ValueError('Please supply either code or '
-                                 'authorization_code parameters.')
+                                 'authorization_response parameters.')
 
 
         body = self._client.prepare_request_body(code=code, body=body,
                 redirect_uri=self.redirect_uri, username=username,
                 password=password, **kwargs)
 
+        client_id = kwargs.get('client_id', '')
+        if auth is None:
+            if client_id:
+                log.debug('Encoding client_id "%s" with client_secret as Basic auth credentials.', client_id)
+                client_secret = kwargs.get('client_secret', '')
+                client_secret = client_secret if client_secret is not None else ''
+                auth = requests.auth.HTTPBasicAuth(client_id, client_secret)
+            elif username:
+                if password is None:
+                    raise ValueError('Username was supplied, but not password.')
+                log.debug('Encoding username, password as Basic auth credentials.')
+                auth = requests.auth.HTTPBasicAuth(username, password)
+
+        headers = headers or {
+            'Accept': 'application/json',
+            'Content-Type': 'application/x-www-form-urlencoded;charset=UTF-8',
+        }
+        self.token = {}
         if method.upper() == 'POST':
             r = self.post(token_url, data=dict(urldecode(body)),
-                headers={'Accept': 'application/json'}, auth=auth,
-                verify=verify)
+                timeout=timeout, headers=headers, auth=auth,
+                verify=verify, proxies=proxies)
             log.debug('Prepared fetch token request body %s', body)
         elif method.upper() == 'GET':
             # if method is not 'POST', switch body to querystring and GET
             r = self.get(token_url, params=dict(urldecode(body)),
-                headers={'Accept': 'application/json'}, auth=auth,
-                verify=verify)
+                timeout=timeout, headers=headers, auth=auth,
+                verify=verify, proxies=proxies)
             log.debug('Prepared fetch token request querystring %s', body)
         else:
             raise ValueError('The method kwarg must be POST or GET.')
 
         log.debug('Request to fetch token completed with status %s.',
                   r.status_code)
+        log.debug('Request headers were %s', r.request.headers)
+        log.debug('Request body was %s', r.request.body)
         log.debug('Response headers were %s and content %s.',
                   r.headers, r.text)
         log.debug('Invoking %d token response hooks.',
@@ -194,7 +258,7 @@ class OAuth2Session(requests.Session):
         return self.token
 
     def refresh_token(self, token_url, refresh_token=None, body='', auth=None,
-                      **kwargs):
+                      timeout=None, headers=None, verify=True, proxies=None, **kwargs):
         """Fetch a new access token using a refresh token.
 
         :param token_url: The token endpoint, must be HTTPS.
@@ -202,6 +266,8 @@ class OAuth2Session(requests.Session):
         :param body: Optional application/x-www-form-urlencoded body to add the
                      include in the token request. Prefer kwargs over body.
         :param auth: An auth tuple or method as accepted by requests.
+        :param timeout: Timeout of the request in seconds.
+        :param verify: Verify SSL certificate.
         :param kwargs: Extra parameters to include in the token request.
         :return: A token dict
         """
@@ -211,9 +277,7 @@ class OAuth2Session(requests.Session):
         if not is_secure_transport(token_url):
             raise InsecureTransportError()
 
-        # Need to nullify token to prevent it from being added to the request
         refresh_token = refresh_token or self.token.get('refresh_token')
-        self.token = {}
 
         log.debug('Adding auto refresh key word arguments %s.',
                   self.auto_refresh_kwargs)
@@ -221,7 +285,17 @@ class OAuth2Session(requests.Session):
         body = self._client.prepare_refresh_body(body=body,
                 refresh_token=refresh_token, scope=self.scope, **kwargs)
         log.debug('Prepared refresh token request body %s', body)
-        r = self.post(token_url, data=dict(urldecode(body)), auth=auth)
+
+        if headers is None:
+            headers = {
+                'Accept': 'application/json',
+                'Content-Type': (
+                    'application/x-www-form-urlencoded;charset=UTF-8'
+                ),
+            }
+
+        r = self.post(token_url, data=dict(urldecode(body)), auth=auth,
+            timeout=timeout, headers=headers, verify=verify, withhold_token=True, proxies=proxies)
         log.debug('Request to refresh token completed with status %s.',
                   r.status_code)
         log.debug('Response headers were %s and content %s.',
@@ -238,11 +312,12 @@ class OAuth2Session(requests.Session):
             self.token['refresh_token'] = refresh_token
         return self.token
 
-    def request(self, method, url, data=None, headers=None, **kwargs):
+    def request(self, method, url, data=None, headers=None, withhold_token=False,
+                client_id=None, client_secret=None, **kwargs):
         """Intercept all requests and add the OAuth 2 token if present."""
         if not is_secure_transport(url):
             raise InsecureTransportError()
-        if self.token:
+        if self.token and not withhold_token:
             log.debug('Invoking %d protected resource request hooks.',
                       len(self.compliance_hook['protected_request']))
             for hook in self.compliance_hook['protected_request']:
@@ -258,7 +333,15 @@ class OAuth2Session(requests.Session):
                 if self.auto_refresh_url:
                     log.debug('Auto refresh is set, attempting to refresh at %s.',
                               self.auto_refresh_url)
-                    token = self.refresh_token(self.auto_refresh_url)
+
+                    # We mustn't pass auth twice.
+                    auth = kwargs.pop('auth', None)
+                    if client_id and client_secret and (auth is None):
+                        log.debug('Encoding client_id "%s" with client_secret as Basic auth credentials.', client_id)
+                        auth = requests.auth.HTTPBasicAuth(client_id, client_secret)
+                    token = self.refresh_token(
+                        self.auto_refresh_url, auth=auth, **kwargs
+                    )
                     if self.token_updater:
                         log.debug('Updating token to %s using %s.',
                                   token, self.token_updater)
